from django.contrib import admin

from kmuhelper.modules.api import admin as apiadmin
from kmuhelper.modules.emails import admin as emailadmin
from kmuhelper.modules.main import admin as mainadmin
from kmuhelper.modules.app import admin as appadmin
from kmuhelper.modules.settings import admin as settingsadmin

from kmuhelper.modules.integrations.paymentimport import admin as paymentimportadmin

# Disable "view on site" globally (for other apps too!)

admin.site.site_url = None

# List of all models and their corresponding admins

modeladmins = []

for adminclass in [
    apiadmin,
    emailadmin,
    mainadmin,
    appadmin,
    settingsadmin,
    paymentimportadmin,
]:
    modeladmins.extend(adminclass.modeladmins)



from django.contrib import admin

from kmuhelper.modules.app.models import (
    App_ToDo,
    App_Shipping,
    App_IncomingPayments,
    App_Stock,
    App_Arrival,
)
from kmuhelper.modules.main.admin import (
    NoteAdmin,
    OrderAdmin,
    SupplyAdmin,
    ProductAdmin,
)
from kmuhelper.overrides import CustomModelAdmin


#######


class App_AdminBase(CustomModelAdmin):
    hidden = True


@admin.register(App_ToDo)
class App_ToDoAdmin(App_AdminBase, NoteAdmin):
    list_editable = ["priority", "done"]
    list_filter = ["priority"]

    ordering = ["-priority", "created_at"]


@admin.register(App_Shipping)
class App_ShippingAdmin(App_AdminBase, OrderAdmin):
    list_display = (
        "id",
        "info",
        "status",
        "shipped_on",
        "is_shipped",
        "tracking_number",
        "linked_note_html",
    )
    list_editable = ("shipped_on", "is_shipped", "status", "tracking_number")
    list_filter = ("status", "is_paid")

    ordering = ("is_paid", "-date")

    actions = ()


@admin.register(App_IncomingPayments)
class App_IncomingPaymentsAdmin(App_AdminBase, OrderAdmin):
    list_display = (
        "id",
        "info",
        "status",
        "paid_on",
        "is_paid",
        "display_cached_sum",
        "display_payment_conditions",
        "linked_note_html",
    )
    list_editable = ("paid_on", "is_paid", "status")
    list_filter = ("status", "is_shipped", "payment_method")

    ordering = (
        "status",
        "invoice_date",
    )

    actions = ()


@admin.register(App_Stock)
class App_StockAdmin(App_AdminBase, ProductAdmin):
    list_display = (
        "nr",
        "clean_name",
        "stock_current",
        "get_current_price",
        "note",
        "linked_note_html",
    )
    list_display_links = ("nr",)
    list_editable = ["stock_current"]

    actions = ["reset_stock"]


@admin.register(App_Arrival)
class App_ArrivalAdmin(App_AdminBase, SupplyAdmin):
    list_display = ("name", "date", "total_quantity", "linked_note_html")
    list_filter = ()


#


modeladmins = [
    (App_ToDo, App_ToDoAdmin),
    (App_Shipping, App_ShippingAdmin),
    (App_IncomingPayments, App_IncomingPaymentsAdmin),
    (App_Stock, App_StockAdmin),
    (App_Arrival, App_ArrivalAdmin),
]



import json
from random import randint
from urllib.parse import urlencode

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy, gettext, ngettext
from django.views.decorators.csrf import csrf_exempt
from rich import print

from kmuhelper import settings
from kmuhelper.decorators import (
    require_object,
    require_all_kmuhelper_perms,
    require_any_kmuhelper_perms,
)
from kmuhelper.modules.integrations.woocommerce.api import WooCommerce
from kmuhelper.modules.integrations.woocommerce.forms import WooCommerceSettingsForm
from kmuhelper.modules.integrations.woocommerce.utils import (
    is_connected,
    base64_hmac_sha256,
    random_secret,
    test_wc_url,
)
from kmuhelper.modules.main.models import Product, Customer, ProductCategory, Order
from kmuhelper.utils import render_error

_ = gettext_lazy


def log(string, *args):
    print("[deep_pink4][KMUHelper][/] -", string, *args)


NOT_CONNECTED_ERRMSG = _("WooCommerce wurde scheinbar nicht richtig verbunden!")

#####


@csrf_exempt
def wc_auth_key(request):
    """Endpoint for receiving the keys from WooCommerce"""

    try:
        data = json.loads(request.body)
        storeurl = request.headers.get("user-agent").split(";")[1].lstrip()
        savedurl = settings.get_db_setting("wc-url")

        storedomain = storeurl.lstrip("https://").lstrip("http://").split("/")[0]
        saveddomain = savedurl.lstrip("https://").lstrip("http://").split("/")[0]

        if saveddomain == storedomain:
            settings.set_secret_db_setting("wc-url", storeurl)
            settings.set_secret_db_setting("wc-consumer_key", data["consumer_key"])
            settings.set_secret_db_setting(
                "wc-consumer_secret", data["consumer_secret"]
            )

            return JsonResponse({"success": True}, status=200)

        return JsonResponse({"success": False, "reason": "Unknown URL!"}, status=401)
    except KeyError:
        return JsonResponse({"success": False, "reason": "Useless data!"}, status=400)


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_any_kmuhelper_perms()
def wc_auth_end(request):
    if request.GET.get("success") == "1":
        messages.success(request, gettext("WooCommerce erfolgreich verbunden!"))
    else:
        messages.error(request, gettext("WooCommerce konnte nicht verbunden werden!"))
    return redirect(reverse("kmuhelper:wc-settings"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_setting"])
def wc_auth_start(request):
    shopurl = settings.get_db_setting("wc-url", "Bestätigt")

    if not shopurl or not shopurl.startswith("http"):
        messages.error(
            request,
            gettext("Please enter a valid WooCommerce URL, beginning with 'https://'!"),
        )
        return redirect(reverse("kmuhelper:wc-settings"))

    kmuhelperurl = request.get_host()
    params = {
        "app_name": "KMUHelper",
        "scope": "read",
        "user_id": randint(100000, 999999),
        "return_url": kmuhelperurl + reverse("kmuhelper:wc-auth-end"),
        "callback_url": kmuhelperurl + reverse("kmuhelper:wc-auth-key"),
    }
    query_string = urlencode(params)

    url = "%s%s?%s" % (shopurl, "/wc-auth/v1/authorize", query_string)
    return redirect(url)


# Note: Change permission is required because the view redirects to the settings page
@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_setting"])
def wc_system_status(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        status = WooCommerce.get_system_status()

        if status:
            messages.success(request, _("WooCommerce is connected and works!"))
        else:
            messages.error(request, _("WooCommerce doesn't seem to work!"))
    return redirect(reverse("kmuhelper:wc-settings"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["add_product"])
def wc_import_products(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        count = WooCommerce.product_import()
        messages.success(
            request,
            ngettext(
                "%d new product has been imported from WooCommerce!",
                "%d new products have been imported from WooCommerce!",
                count,
            )
            % count,
        )
    return redirect(reverse("admin:kmuhelper_product_changelist"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["add_customer"])
def wc_import_customers(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        count = WooCommerce.customer_import()
        messages.success(
            request,
            ngettext(
                "%d new customer has been imported from WooCommerce!",
                "%d new customers have been imported from WooCommerce!",
                count,
            )
            % count,
        )
    return redirect(reverse("admin:kmuhelper_customer_changelist"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["add_productcategory"])
def wc_import_categories(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        count = WooCommerce.category_import()
        messages.success(
            request,
            ngettext(
                "%d new product category has been imported from WooCommerce!",
                "%d new product categories have been imported from WooCommerce!",
                count,
            )
            % count,
        )
    return redirect(reverse("admin:kmuhelper_productcategory_changelist"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["add_order"])
def wc_import_orders(request):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        count = WooCommerce.order_import()
        messages.success(
            request,
            ngettext(
                "%d new order has been imported from WooCommerce!",
                "%d new orders have been imported from WooCommerce!",
                count,
            )
            % count,
        )
    return redirect(reverse("admin:kmuhelper_order_changelist"))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_product"])
@require_object(Product)
def wc_update_product(request, obj):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        product = WooCommerce.product_update(obj)
        messages.success(request, gettext("Product '%s' updated!") % str(product))
    return redirect(reverse("admin:kmuhelper_product_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_customer"])
@require_object(Customer)
def wc_update_customer(request, obj):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        customer = WooCommerce.customer_update(obj)
        messages.success(request, gettext("Customer '%s' updated!") % str(customer))
    return redirect(reverse("admin:kmuhelper_customer_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_productcategory"])
@require_object(ProductCategory)
def wc_update_category(request, obj):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        category = WooCommerce.category_update(obj)
        messages.success(
            request, gettext("Product category '%s' updated!") % str(category)
        )
    return redirect(reverse("admin:kmuhelper_productcategory_change", args=[obj.pk]))


@login_required(login_url=reverse_lazy("kmuhelper:login"))
@require_all_kmuhelper_perms(["change_order"])
@require_object(Order)
def wc_update_order(request, obj):
    if not is_connected():
        messages.error(request, NOT_CONNECTED_ERRMSG)
    else:
        order = WooCommerce.order_update(obj)
        messages.success(request, gettext("Order '%s' updated!") % str(order))
    return redirect(reverse("admin:kmuhelper_order_change", args=[obj.pk]))


@csrf_exempt
def wc_webhooks(request):
    "Endpoint for WooCommerce webhooks"

    # 1. Check request method

    if request.method != "POST":
        messages.warning(
            request,
            gettext(
                "This endpoint is only available via POST and not meant to be used by humans!"
            ),
        )
        return render_error(request, status=405)

    # 2. Check request source

    if not (
        "x-wc-webhook-topic" in request.headers
        and "x-wc-webhook-source" in request.headers
    ):
        log(
            "[orange_red1]WooCommerce Webhook accepted but ignored (no usable headers)![/] "
            + "(Might be a test request from WooCommerce)"
        )
        return JsonResponse(
            {
                "accepted": True,
                "info": "Request was accepted but ignored because it doesn't contain any usable info!",
            },
            status=200,
        )

    stored_url = (
        settings.get_secret_db_setting("wc-url")
        .lstrip("https://")
        .lstrip("http://")
        .split("/")[0]
    )
    received_url = (
        request.headers["x-wc-webhook-source"]
        .lstrip("https://")
        .lstrip("http://")
        .split("/")[0]
    )

    if not received_url == stored_url:
        log(
            "[orange_red1]WooCommerce Webhook rejected (unexpected domain)![/] "
            + "- Expected:",
            stored_url,
            "- Received:",
            received_url,
        )
        return JsonResponse(
            {
                "accepted": False,
                "reason": "Unknown domain!",
            },
            status=403,
        )

    # 3. Check request signature

    if not ("x-wc-webhook-signature" in request.headers):
        log("[orange_red1]WooCommerce Webhook rejected (no signature)![/]")
        return JsonResponse(
            {
                "accepted": False,
                "reason": "Invalid signature!",
            },
            status=403,
        )

    secret = settings.get_db_setting("wc-webhook-secret")

    if secret:
        received_signature = bytes(
            request.headers["x-wc-webhook-signature"], encoding="utf-8"
        )
        expected_signature = base64_hmac_sha256(str(secret).encode(), request.body)

        if received_signature == expected_signature:
            log("[green]WooCommerce Webhook signature check successful![/]")
        else:
            log(
                "[orange_red1]WooCommerce Webhook signature check failed![/] "
                + "- Expected:",
                expected_signature,
                "- Received:",
                received_signature,
            )
            return JsonResponse(
                {
                    "accepted": False,
                    "reason": "Signature mismatch!",
                },
                status=403,
            )

    else:
        log(
            "[orange_red1]Skipped WooCommerce Webhook signature check (no secret available)![/]"
        )

    # Do stuff

    log("WooCommerce Webhook is being processed...")

    topic = request.headers["x-wc-webhook-topic"]
    obj = json.loads(request.body)
    if topic in ("product.updated", "product.created"):
        if Product.objects.filter(woocommerceid=obj["id"]).exists():
            WooCommerce.product_update(
                Product.objects.get(woocommerceid=obj["id"]), obj
            )
        else:
            WooCommerce.product_create(obj)
    elif topic == "product.deleted":
        if Product.objects.filter(woocommerceid=obj["id"]).exists():
            product = Product.objects.get(woocommerceid=obj["id"])
            product.woocommerceid = 0
            product.save()
    elif topic in ("customer.updated", "customer.created"):
        if Customer.objects.filter(woocommerceid=obj["id"]).exists():
            WooCommerce.customer_update(
                Customer.objects.get(woocommerceid=obj["id"]), obj
            )
        else:
            WooCommerce.customer_create(obj)
    elif topic == "customer.deleted":
        if Customer.objects.filter(woocommerceid=obj["id"]).exists():
            customer = Customer.objects.get(woocommerceid=obj["id"])
            customer.woocommerceid = 0
            customer.save()
    elif topic in ("order.updated", "order.created"):
        if Order.objects.filter(woocommerceid=obj["id"]).exists():
            WooCommerce.order_update(Order.objects.get(woocommerceid=obj["id"]), obj)
        else:
            WooCommerce.order_create(obj)
    elif topic == "order.deleted":
        if Order.objects.filter(woocommerceid=obj["id"]).exists():
            order = Order.objects.get(woocommerceid=obj["id"])
            order.woocommerceid = 0
            order.save()
    else:
        log(f"[orange_red1]Unbekanntes Thema: '{topic}'")

    return JsonResponse({"accepted": True})


# Settings


@login_required(login_url=reverse_lazy("admin:login"))
@require_all_kmuhelper_perms(["change_setting"])
def wc_settings(request):
    if request.method == "POST":
        form = WooCommerceSettingsForm(request.POST)
        if form.is_valid():
            form.save_settings()
            messages.success(request, gettext("Einstellungen gespeichert!"))
            # Redirect to prevent resending the form on reload
            return redirect("kmuhelper:wc-settings")
    else:
        form = WooCommerceSettingsForm()

    secreturl = settings.get_secret_db_setting("wc-url", None)
    url = settings.get_db_setting("wc-url", None)

    return render(
        request,
        "kmuhelper/integrations/woocommerce/settings.html",
        {
            "form": form,
            "has_permission": True,
            "random_secret": random_secret(),
            "kmuhelper_url": request.get_host(),
            "wc_url": secreturl,
            "is_connected": is_connected(),
            "is_url_valid": test_wc_url(url),
        },
    )



from django.utils.html import strip_tags
from rich import print
from rich.progress import Progress
from woocommerce import API as WCAPI

from kmuhelper import settings, constants
from kmuhelper.modules.main.models import Product, Customer, ProductCategory, Order
from kmuhelper.utils import runden

PREFIX = "[deep_pink4][KMUHelper WooCommerce][/] -"


def log(string, *args):
    print(PREFIX, string, *args)


def preparestring(string):
    """Prepare a HTML string for import"""
    return (
        strip_tags(string.replace("</p>", " ").replace("</strong>", " "))
        .replace("&#8211;", "-")
        .replace("&#8211;", " ")
        .replace("&#215;", "x")
        .replace("&#8220;", '"')
        .replace("&#8221;", '"')
        .replace("&nbsp;", " ")
        .replace("&lt;", "<")
        .replace("&gt;", ">")
        .replace("&amp;", "&")
        .replace("&quot;", '"')
        .replace("&#8222;", '"')
        .replace("  ", " ")
    )


###############


class WooCommerce:
    """Manage connection with WooCommerce"""

    @classmethod
    def get_api(cls):
        """Create a API object from data stored in settings"""
        return WCAPI(
            url=settings.get_secret_db_setting("wc-url"),
            consumer_key=settings.get_secret_db_setting("wc-consumer_key"),
            consumer_secret=settings.get_secret_db_setting("wc-consumer_secret"),
        )

    # System status

    @classmethod
    def get_system_status(cls):
        """Get system status from WooCommerce"""
        wcapi = cls.get_api()
        try:
            request = wcapi.get("system_status")
            request.raise_for_status()
            return request.json()
        except Exception as e:
            log("Error while getting system status: ", e)
            return False

    # General methods

    @classmethod
    def product_create(cls, product, api=None):
        """Create a new product from WooCommerce data"""
        wcapi = api or cls.get_api()

        try:
            selling_price = float(product["price"])
        except ValueError:
            selling_price = 0.0

        newproduct = Product.objects.create(
            woocommerceid=product["id"],
            article_number=product["sku"],
            name=preparestring(product["name"]),
            short_description=preparestring(product["short_description"]),
            description=preparestring(product["description"]),
            selling_price=selling_price,
            image_url=(product["images"][0]["src"])
            if len(product["images"]) > 0
            else "",
            sale_from=(
                product["date_on_sale_from_gmt"] + "+00:00"
                if product["date_on_sale_from"]
                else None
            ),
            sale_to=(
                product["date_on_sale_to_gmt"] + "+00:00"
                if product["date_on_sale_to"]
                else None
            ),
            sale_price=(product["sale_price"] if product["sale_price"] else None),
        )
        if product["manage_stock"]:
            newproduct.stock_current = product["stock_quantity"]
        for category in product["categories"]:
            obj, created = ProductCategory.objects.get_or_create(
                woocommerceid=category["id"]
            )
            if created:
                cls.category_update(obj, api=wcapi)
            newproduct.categories.add(obj)
        newproduct.save()
        log("Product created:", str(newproduct))
        return newproduct

    @classmethod
    def product_update(cls, product, newproduct=None, api=None):
        """Update a specific product from WooCommerce"""
        wcapi = api or cls.get_api()

        if not newproduct:
            newproduct = wcapi.get(f"products/{product.woocommerceid}").json()

        try:
            product.selling_price = float(newproduct["price"])
        except ValueError:
            pass
        except KeyError:
            if (
                "code" in newproduct
                and newproduct["code"] == "woocommerce_rest_product_invalid_id"
            ):
                log(
                    "[red]Product does not exist in WooCommerce![/] Link removed!",
                    str(product),
                )
                product.woocommerceid = 0
                product.save()
                return product

        product.article_number = newproduct["sku"]
        product.name = preparestring(newproduct["name"])
        product.short_description = preparestring(newproduct["short_description"])
        product.description = preparestring(newproduct["description"])
        product.image_url = (
            (newproduct["images"][0]["src"]) if len(newproduct["images"]) > 0 else ""
        )
        if newproduct["date_on_sale_from_gmt"]:
            product.sale_from = newproduct["date_on_sale_from_gmt"] + "+00:00"
        if newproduct["date_on_sale_to_gmt"]:
            product.sale_to = newproduct["date_on_sale_to_gmt"] + "+00:00"
        if newproduct["sale_price"]:
            product.sale_price = newproduct["sale_price"]

        product.categories.clear()
        newcategories = []
        for category in newproduct["categories"]:
            obj, created = ProductCategory.objects.get_or_create(
                woocommerceid=category["id"]
            )
            if created:
                cls.category_update(obj, api=wcapi)
            newcategories.append(obj)
        product.categories.add(*newcategories)

        product.save()
        log("Product updated:", str(product))
        return product

    @classmethod
    def product_import(cls, api=None):
        """Import new products from WooCommerce"""
        wcapi = api or cls.get_api()

        with Progress() as progress:
            task_prepare = progress.add_task(
                PREFIX + " [orange_red1]Preparing product download...", total=1
            )

            excludeids = ",".join(
                [
                    str(obj.woocommerceid)
                    for obj in Product.objects.all().exclude(woocommerceid=0)
                ]
            )
            productlist = []

            progress.update(task_prepare, advance=1)
            progress.stop_task(task_prepare)

            task_download = progress.add_task(
                PREFIX + " [green]Downloading products..."
            )

            r = wcapi.get("products?exclude=" + excludeids)
            productlist += r.json()

            progress.update(
                task_download, advance=1, total=int(r.headers["X-WP-TotalPages"])
            )

            for page in range(2, int(r.headers["X-WP-TotalPages"]) + 1):
                productlist += wcapi.get(
                    "products?exclude=" + excludeids + "&page=" + str(page)
                ).json()
                progress.update(task_download, advance=1)

            progress.stop_task(task_download)

            if productlist:
                task_process = progress.add_task(
                    PREFIX + " [cyan]Processing products...", total=len(productlist)
                )
                for product in productlist:
                    cls.product_create(product, api=wcapi)
                    progress.update(task_process, advance=1)
                progress.stop_task(task_process)
        return len(productlist)

    @classmethod
    def product_bulk_update(cls, products, api=None):
        """Update every product in a queryset"""
        wcapi = api or cls.get_api()

        with Progress() as progress:
            task = progress.add_task(
                PREFIX + " [orange_red1]Updating products...", total=products.count()
            )
            successcount = 0
            errorcount = 0
            for product in products:
                if product.woocommerceid:
                    cls.product_update(product, api=wcapi)
                    successcount += 1
                else:
                    errorcount += 1
                progress.update(task, advance=1)
            progress.stop_task(task)
        return (successcount, errorcount)

    @classmethod
    def customer_create(cls, customer):
        """Create a new customer from WooCommerce data"""
        newcustomer = Customer.objects.create(
            woocommerceid=customer["id"],
            email=customer["email"],
            first_name=customer["first_name"],
            last_name=customer["last_name"],
            company=customer["billing"]["company"],
            username=customer["username"],
            avatar_url=customer["avatar_url"],
            addr_billing_first_name=customer["billing"]["first_name"],
            addr_billing_last_name=customer["billing"]["last_name"],
            addr_billing_company=customer["billing"]["company"],
            addr_billing_address_1=customer["billing"]["address_1"],
            addr_billing_address_2=customer["billing"]["address_2"],
            addr_billing_city=customer["billing"]["city"],
            addr_billing_state=customer["billing"]["state"],
            addr_billing_postcode=customer["billing"]["postcode"],
            addr_billing_country=customer["billing"]["country"],
            addr_billing_email=customer["billing"]["email"],
            addr_billing_phone=customer["billing"]["phone"],
            addr_shipping_first_name=customer["shipping"]["first_name"],
            addr_shipping_last_name=customer["shipping"]["last_name"],
            addr_shipping_company=customer["shipping"]["company"],
            addr_shipping_address_1=customer["shipping"]["address_1"],
            addr_shipping_address_2=customer["shipping"]["address_2"],
            addr_shipping_city=customer["shipping"]["city"],
            addr_shipping_state=customer["shipping"]["state"],
            addr_shipping_postcode=customer["shipping"]["postcode"],
            addr_shipping_country=customer["shipping"]["country"],
            addr_shipping_email="",
            addr_shipping_phone="",
        )
        log("Customer created:", str(newcustomer))
        return newcustomer

    @classmethod
    def customer_update(cls, customer, newcustomer=None, api=None):
        """Update a specific customer from WooCommerce"""
        wcapi = api or cls.get_api()

        if not newcustomer:
            newcustomer = wcapi.get(f"customers/{customer.woocommerceid}").json()

        if (
            "code" in newcustomer
            and newcustomer["code"] == "woocommerce_rest_customer_invalid_id"
        ):
            log(
                "[red]Customer does not exist in WooCommerce![/] Link removed.",
                str(customer),
            )
            customer.woocommerceid = 0
            customer.save()
            return customer

        customer.email = newcustomer["email"]
        customer.first_name = newcustomer["first_name"]
        customer.last_name = newcustomer["last_name"]
        customer.company = (
            newcustomer["billing"]["company"]
            if newcustomer["billing"]["company"]
            else customer.company
        )
        customer.username = newcustomer["username"]
        customer.avatar_url = newcustomer["avatar_url"]

        customer.addr_billing_first_name = newcustomer["billing"]["first_name"]
        customer.addr_billing_last_name = newcustomer["billing"]["last_name"]
        customer.addr_billing_company = newcustomer["billing"]["company"]
        customer.addr_billing_address_1 = newcustomer["billing"]["address_1"]
        customer.addr_billing_address_2 = newcustomer["billing"]["address_2"]
        customer.addr_billing_city = newcustomer["billing"]["city"]
        customer.addr_billing_state = newcustomer["billing"]["state"]
        customer.addr_billing_postcode = newcustomer["billing"]["postcode"]
        customer.addr_billing_country = newcustomer["billing"]["country"]
        customer.addr_billing_email = newcustomer["billing"]["email"]
        customer.addr_billing_phone = newcustomer["billing"]["phone"]

        customer.addr_shipping_first_name = newcustomer["shipping"]["first_name"]
        customer.addr_shipping_last_name = newcustomer["shipping"]["last_name"]
        customer.addr_shipping_company = newcustomer["shipping"]["company"]
        customer.addr_shipping_address_1 = newcustomer["shipping"]["address_1"]
        customer.addr_shipping_address_2 = newcustomer["shipping"]["address_2"]
        customer.addr_shipping_city = newcustomer["shipping"]["city"]
        customer.addr_shipping_state = newcustomer["shipping"]["state"]
        customer.addr_shipping_postcode = newcustomer["shipping"]["postcode"]
        customer.addr_shipping_country = newcustomer["shipping"]["country"]
        customer.save()
        log("Customer updated:", str(customer))
        return customer

    @classmethod
    def customer_import(cls, api=None):
        """Import new customers from WooCommerce"""
        wcapi = api or cls.get_api()

        with Progress() as progress:
            task_prepare = progress.add_task(
                PREFIX + " [orange_red1]Preparing customer download...", total=1
            )

            excludeids = ",".join(
                [
                    str(obj.woocommerceid)
                    for obj in Customer.objects.all().exclude(woocommerceid=0)
                ]
            )
            customerlist = []

            progress.update(task_prepare, advance=1)
            progress.stop_task(task_prepare)

            task_download = progress.add_task(
                PREFIX + " [green]Downloading customers..."
            )

            r = wcapi.get("customers?exclude=" + excludeids)
            customerlist += r.json()

            progress.update(
                task_download, advance=1, total=int(r.headers["X-WP-TotalPages"])
            )

            for page in range(2, int(r.headers["X-WP-TotalPages"]) + 1):
                customerlist += wcapi.get(
                    "customers?exclude=" + excludeids + "&page=" + str(page)
                ).json()
                progress.update(task_download, advance=1)
            progress.stop_task(task_download)

            if customerlist:
                task_process = progress.add_task(
                    PREFIX + " [cyan]Processing customers...", total=len(customerlist)
                )
                for customer in customerlist:
                    cls.customer_create(customer)
                    progress.update(task_process, advance=1)
                progress.stop_task(task_process)
        return len(customerlist)

    @classmethod
    def customer_bulk_update(cls, customers, api=None):
        """Update every customer in a queryset"""
        wcapi = api or cls.get_api()

        with Progress() as progress:
            task = progress.add_task(
                PREFIX + " [orange_red1]Updating customers...", total=customers.count()
            )
            successcount = 0
            errorcount = 0
            for customer in customers:
                if customer.woocommerceid:
                    cls.customer_update(customer, api=wcapi)
                    successcount += 1
                else:
                    errorcount += 1
                progress.update(task, advance=1)
            progress.stop_task(task)
        return (successcount, errorcount)

    @classmethod
    def category_update(cls, category, newcategory=None, api=None):
        """Update a specific category from WooCommerce"""
        wcapi = api or cls.get_api()

        if not newcategory:
            newcategory = wcapi.get(
                f"products/categories/{category.woocommerceid}"
            ).json()

        if (
            "code" in newcategory
            and newcategory["code"] == "woocommerce_rest_category_invalid_id"
        ):
            log(
                "[red]Category does not exist in WooCommerce![/] Link removed.",
                str(category),
            )
            category.woocommerceid = 0
            category.save()
            return category

        category.name = preparestring(newcategory["name"])
        category.description = preparestring(newcategory["description"])
        category.image_url = (
            (newcategory["image"]["src"]) if newcategory["image"] else ""
        )
        if newcategory["parent"]:
            obj, created = ProductCategory.objects.get_or_create(
                woocommerceid=newcategory["parent"]
            )
            if created:
                cls.category_update(obj, api=wcapi)
            category.parent_category = obj
        category.save()
        log("Category updated:", str(category))
        return category

    @classmethod
    def category_import(cls, api=None):
        """Import new categories from WooCommerce"""
        wcapi = api or cls.get_api()

        with Progress() as progress:
            task_prepare = progress.add_task(
                PREFIX + " [orange_red1]Preparing category download...", total=1
            )

            excludeids = ",".join(
                [
                    str(obj.woocommerceid)
                    for obj in ProductCategory.objects.all().exclude(woocommerceid=0)
                ]
            )
            categorylist = []

            progress.update(task_prepare, advance=1)
            progress.stop_task(task_prepare)

            task_download = progress.add_task(
                PREFIX + " [green]Downloading categories..."
            )

            r = wcapi.get("products/categories?exclude=" + excludeids)
            categorylist += r.json()

            progress.update(
                task_download, advance=1, total=int(r.headers["X-WP-TotalPages"])
            )

            for page in range(2, int(r.headers["X-WP-TotalPages"]) + 1):
                categorylist += wcapi.get(
                    "products/categories?exclude=" + excludeids + "&page=" + str(page)
                ).json()
                progress.update(task_download, advance=1)

            progress.stop_task(task_download)

            categorieswithparents = []

            if categorylist:
                task_process = progress.add_task(
                    PREFIX + " [cyan]Processing categories...", total=len(categorylist)
                )
                for category in categorylist:
                    newcategory = ProductCategory.objects.create(
                        name=preparestring(category["name"]),
                        description=preparestring(category["description"]),
                        image_url=(category["image"]["src"])
                        if category["image"]
                        else "",
                        woocommerceid=category["id"],
                    )
                    log("Category created:", category["name"])
                    if category["parent"]:
                        categorieswithparents.append((newcategory, category["parent"]))
                    progress.update(task_process, advance=1)
                progress.stop_task(task_process)

            if categorieswithparents:
                task_process2 = progress.add_task(
                    PREFIX + " [cyan]Processing category dependencies...",
                    total=len(categorieswithparents),
                )
                for cat, parentwcid in categorieswithparents:
                    cat.parent_category = ProductCategory.objects.get(
                        woocommerceid=parentwcid
                    )
                    cat.save()
                    progress.update(task_process2, advance=1)
                progress.stop_task(task_process2)
        return len(categorylist)

    @classmethod
    def category_bulk_update(cls, categories, api=None):
        """Update every category in a queryset"""
        wcapi = api or cls.get_api()

        with Progress() as progress:
            task = progress.add_task(
                PREFIX + " [orange_red1]Updating categories...",
                total=categories.count(),
            )
            successcount = 0
            errorcount = 0
            for category in categories:
                if category.woocommerceid:
                    cls.category_update(category, api=wcapi)
                    successcount += 1
                else:
                    errorcount += 1
                progress.update(task, advance=1)
            progress.stop_task(task)
        return (successcount, errorcount)

    @classmethod
    def order_create(cls, order, api=None, sendstockwarning=True):
        """Create a new order from WooCommerce data"""
        wcapi = api or cls.get_api()

        neworder = Order.objects.create(
            woocommerceid=order["id"],
            status=order["status"],
            is_shipped=(True if order["status"] == "completed" else False),
            is_removed_from_stock=(True if order["status"] == "completed" else False),
            payment_method=order["payment_method"],
            is_paid=(True if order["date_paid"] else False),
            customer_note=order["customer_note"],
            order_key=order["order_key"],
            addr_billing_first_name=order["billing"]["first_name"],
            addr_billing_last_name=order["billing"]["last_name"],
            addr_billing_company=order["billing"]["company"],
            addr_billing_address_1=order["billing"]["address_1"],
            addr_billing_address_2=order["billing"]["address_2"],
            addr_billing_city=order["billing"]["city"],
            addr_billing_state=order["billing"]["state"],
            addr_billing_postcode=order["billing"]["postcode"],
            addr_billing_country=order["billing"]["country"],
            addr_billing_email=order["billing"]["email"],
            addr_billing_phone=order["billing"]["phone"],
            addr_shipping_first_name=order["shipping"]["first_name"],
            addr_shipping_last_name=order["shipping"]["last_name"],
            addr_shipping_company=order["shipping"]["company"],
            addr_shipping_address_1=order["shipping"]["address_1"],
            addr_shipping_address_2=order["shipping"]["address_2"],
            addr_shipping_city=order["shipping"]["city"],
            addr_shipping_state=order["shipping"]["state"],
            addr_shipping_postcode=order["shipping"]["postcode"],
            addr_shipping_country=order["shipping"]["country"],
        )
        neworder.date = order["date_created_gmt"] + "+00:00"
        if order["customer_id"]:
            customer, created = Customer.objects.get_or_create(
                woocommerceid=int(order["customer_id"])
            )
            if created:
                customer = cls.customer_update(customer, api=wcapi)

            neworder.customer = customer
            neworder.addr_shipping_email = customer.addr_shipping_email
            neworder.addr_shipping_phone = customer.addr_shipping_phone

        for item in order["line_items"]:
            product, created = Product.objects.get_or_create(
                woocommerceid=int(item["product_id"])
            )
            if created:
                product = cls.product_update(product, api=wcapi)

            neworder.products.add(
                product,
                through_defaults={
                    "quantity": int(item["quantity"]),
                    "product_price": runden(float(item["price"])),
                },
            )
        for item in order["shipping_lines"]:
            neworder.fees.through.objects.create(
                order=neworder,
                name=item["method_title"],
                price=float(item["total"]),
                vat_rate=(
                    constants.VAT_RATE_DEFAULT if float(item["total_tax"]) > 0 else 0
                ),
            )
        neworder.save()
        neworder.second_save()
        log("Order created:", str(neworder))

        if sendstockwarning:
            neworder.email_stock_warning()
        return neworder

    @classmethod
    def order_update(cls, order, neworder=None, api=None):
        """Update a specific order from WooCommerce"""
        wcapi = api or cls.get_api()

        if not neworder:
            neworder = wcapi.get(f"orders/{order.woocommerceid}").json()

        if (
            "code" in neworder
            and neworder["code"] == "woocommerce_rest_order_invalid_id"
        ):
            log(
                "[red]Order does not exist in WooCommerce![/] Link removed.", str(order)
            )
            order.woocommerceid = 0
            order.save()
            return order

        if neworder["date_paid"]:
            order.is_paid = True
        order.status = neworder["status"]
        order.customer_note = neworder["customer_note"]

        order.addr_billing_first_name = neworder["billing"]["first_name"]
        order.addr_billing_last_name = neworder["billing"]["last_name"]
        order.addr_billing_company = neworder["billing"]["company"]
        order.addr_billing_address_1 = neworder["billing"]["address_1"]
        order.addr_billing_address_2 = neworder["billing"]["address_2"]
        order.addr_billing_city = neworder["billing"]["city"]
        order.addr_billing_state = neworder["billing"]["state"]
        order.addr_billing_postcode = neworder["billing"]["postcode"]
        order.addr_billing_country = neworder["billing"]["country"]
        order.addr_billing_email = neworder["billing"]["email"]
        order.addr_billing_phone = neworder["billing"]["phone"]

        order.addr_shipping_first_name = neworder["shipping"]["first_name"]
        order.addr_shipping_last_name = neworder["shipping"]["last_name"]
        order.addr_shipping_company = neworder["shipping"]["company"]
        order.addr_shipping_address_1 = neworder["shipping"]["address_1"]
        order.addr_shipping_address_2 = neworder["shipping"]["address_2"]
        order.addr_shipping_city = neworder["shipping"]["city"]
        order.addr_shipping_state = neworder["shipping"]["state"]
        order.addr_shipping_postcode = neworder["shipping"]["postcode"]
        order.addr_shipping_country = neworder["shipping"]["country"]
        order.save()
        log("Order updated: ", str(order))
        return order

    @classmethod
    def order_import(cls, api=None):
        """Import new orders from WooCommerce"""
        wcapi = api or cls.get_api()

        with Progress() as progress:
            task_prepare = progress.add_task(
                PREFIX + " [orange_red1]Preparing order download...", total=1
            )

            excludeids = ",".join(
                [
                    str(obj.woocommerceid)
                    for obj in Order.objects.all().exclude(woocommerceid=0)
                ]
            )
            orderlist = []

            progress.update(task_prepare, advance=1)
            progress.stop_task(task_prepare)

            task_download = progress.add_task(PREFIX + " [green]Downloading orders...")

            r = wcapi.get("orders?exclude=" + excludeids)
            orderlist += r.json()

            progress.update(
                task_download, advance=1, total=int(r.headers["X-WP-TotalPages"])
            )

            for page in range(2, int(r.headers["X-WP-TotalPages"]) + 1):
                orderlist += wcapi.get(
                    "orders?exclude=" + excludeids + "&page=" + str(page)
                ).json()
                progress.update(task_download, advance=1)
            progress.stop_task(task_download)

            if orderlist:
                task_process = progress.add_task(
                    PREFIX + " [cyan]Processing orders...", total=len(orderlist)
                )
                for order in orderlist:
                    cls.order_create(order, api=wcapi, sendstockwarning=False)
                    progress.update(task_process, advance=1)
                progress.stop_task(task_process)
        return len(orderlist)

    @classmethod
    def order_bulk_update(cls, orders, api=None):
        """Update every order in a queryset"""
        wcapi = api or cls.get_api()

        with Progress() as progress:
            task = progress.add_task(
                PREFIX + " [orange_red1]Updating orders...", total=orders.count()
            )
            successcount = 0
            errorcount = 0
            for order in orders:
                if order.woocommerceid:
                    cls.order_update(order, api=wcapi)
                    successcount += 1
                else:
                    errorcount += 1
                progress.update(task, advance=1)
            progress.stop_task(task)
        return (successcount, errorcount)



from urllib.parse import urlencode

from django.contrib import admin
from django.db import models
from django.forms.models import model_to_dict
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _, gettext

from kmuhelper import widgets


class CustomModel(models.Model):
    """django.db.models.Model with custom overrides"""

    def pkfill(self, width=6):
        """Get the primary key of this object padded with zeros"""

        return str(self.pk).zfill(width)

    @property
    def to_dict(self):
        """Return the model fields as a dict"""

        exclude = getattr(self.__class__, "DICT_EXCLUDE_FIELDS", list())
        fields = getattr(self.__class__, "DICT_FIELDS", list())

        if fields:
            return model_to_dict(self, fields=fields, exclude=exclude)

        return model_to_dict(self, exclude=exclude)

    @property
    def to_query_string(self):
        """Return the model fields as a query string"""

        return urlencode(self.to_dict)

    @admin.display(description=_("🔗 Notiz"), ordering="linked_note")
    def linked_note_html(self):
        is_app = getattr(self.__class__, "IS_APP_MODEL", False)
        relname = getattr(self.__class__, "NOTE_RELATION", None)

        if relname is None:
            return None

        if hasattr(self, "linked_note"):
            if is_app:
                viewname = "admin:kmuhelper_app_todo_change"
            else:
                viewname = "admin:kmuhelper_note_change"

            link = reverse(viewname, kwargs={"object_id": self.linked_note.pk})
            text = gettext("Notiz ansehen")
        else:
            if is_app:
                viewname = "admin:kmuhelper_app_todo_add"
            else:
                viewname = "admin:kmuhelper_note_add"

            link = reverse(viewname) + f"?from_{relname}={self.pk}"
            text = gettext("Notiz erstellen")
        return format_html('<a target="_blank" href="{}">{}</a>', link, text)

    class Meta:
        abstract = True


class CustomModelAdmin(admin.ModelAdmin):
    """django.contrib.admin.ModelAdmin with custom overrides"""

    formfield_overrides = {models.JSONField: {"widget": widgets.PrettyJSONWidget}}

    def _get_obj_does_not_exist_redirect(self, request, opts, object_id):
        """Redirect to changelist page instead of admin home if object is not found"""

        super()._get_obj_does_not_exist_redirect(request, opts, object_id)
        info = self.model._meta.app_label, self.model._meta.model_name
        return redirect(reverse("admin:%s_%s_changelist" % info))

    def get_actions(self, request):
        """Make some action not always available"""

        actions = super().get_actions(request)
        if "wc_update" in actions:
            from kmuhelper.modules.integrations.woocommerce.utils import is_connected

            if not is_connected():
                del actions["wc_update"]
        return actions

    def has_module_permission(self, request):
        """Add option to hide model in default admin"""

        if getattr(self.__class__, "hidden", False):
            return {}

        return super().has_module_permission(request)

    # Readonly fields

    def get_readonly_fields(self, request, obj=None) -> tuple:
        if "unlock" in request.GET:
            return tuple(self.readonly_fields)
        return tuple(self.readonly_fields) + tuple(
            self.get_additional_readonly_fields(request, obj)
        )

    def get_additional_readonly_fields(self, request, obj=None) -> list:
        return []

    # Permissions

    NO_CHANGE = False

    def has_change_permission(self, request, obj=None):
        return False if self.NO_CHANGE else super().has_change_permission(request, obj)

    NO_ADD = False

    def has_add_permission(self, request):
        return False if self.NO_ADD else super().has_add_permission(request)

    NO_DELETE = False

    def has_delete_permission(self, request, obj=None):
        return False if self.NO_DELETE else super().has_delete_permission(request, obj)

    NO_VIEW = False

    def has_view_permission(self, request, obj=None):
        return False if self.NO_VIEW else super().has_view_permission(request, obj)


class CustomTabularInline(admin.TabularInline):
    """django.contrib.admin.TabularInline with custom overrides"""

    # Readonly fields

    def get_readonly_fields(self, request, obj=None) -> tuple:
        if "unlock" in request.GET:
            return tuple(self.readonly_fields)
        return tuple(self.readonly_fields) + tuple(
            self.get_additional_readonly_fields(request, obj)
        )

    def get_additional_readonly_fields(self, request, obj=None) -> list:
        return []

    # Permissions

    NO_CHANGE = False

    def has_change_permission(self, request, obj=None):
        return False if self.NO_CHANGE else super().has_change_permission(request, obj)

    NO_ADD = False

    def has_add_permission(self, request, obj=None):
        return False if self.NO_ADD else super().has_add_permission(request, obj)

    NO_DELETE = False

    def has_delete_permission(self, request, obj=None):
        return False if self.NO_DELETE else super().has_delete_permission(request, obj)

    NO_VIEW = False

    def has_view_permission(self, request, obj=None):
        return False if self.NO_VIEW else super().has_view_permission(request, obj)


class CustomStackedInline(admin.StackedInline):
    """django.contrib.admin.StackedInline with custom overrides"""

    # Readonly fields

    def get_readonly_fields(self, request, obj=None) -> tuple:
        if "unlock" in request.GET:
            return tuple(self.readonly_fields)
        return tuple(self.readonly_fields) + tuple(
            self.get_additional_readonly_fields(request, obj)
        )

    def get_additional_readonly_fields(self, request, obj=None) -> list:
        return []

    # Permissions

    NO_CHANGE = False

    def has_change_permission(self, request, obj=None):
        return False if self.NO_CHANGE else super().has_change_permission(request, obj)

    NO_ADD = False

    def has_add_permission(self, request, obj=None):
        return False if self.NO_ADD else super().has_add_permission(request, obj)

    NO_DELETE = False

    def has_delete_permission(self, request, obj=None):
        return False if self.NO_DELETE else super().has_delete_permission(request, obj)

    NO_VIEW = False

    def has_view_permission(self, request, obj=None):
        return False if self.NO_VIEW else super().has_view_permission(request, obj)



from reportlab.platypus import Flowable
from reportlab.lib.units import mm
from reportlab.graphics.shapes import Drawing
from reportlab.graphics.barcode import qr
from reportlab.graphics import renderPDF

from django.utils.translation import pgettext

from kmuhelper.utils import formatprice
import kmuhelper.modules.main.models as models


class QRInvoiceFlowable(Flowable):
    "A Flowable that draws a Swiss QR-Invoice."

    def __init__(
        self,
        total,
        address,
        payment_receiver,
        billing_information,
        qr_refernce_number="",
        unstructured_message="",
        add_cut_info=True,
    ):
        self.width = 210
        self.height = 110
        self._fixedWidth = 210
        self._fixedHeight = 110

        self.total: float = total
        self.address: dict = address
        self.payment_receiver: models.PaymentReceiver = payment_receiver
        self.billing_information: str = billing_information
        self.qr_reference_number: str = qr_refernce_number
        self.unstructured_message: str = unstructured_message
        self.add_cut_info: bool = add_cut_info

    @classmethod
    def from_order(cls, order, add_cut_lines=True):
        elem = cls(
            total=order.cached_sum,
            address=order.addr_billing,
            payment_receiver=order.payment_receiver,
            billing_information=order.get_qr_billing_information(),
            qr_refernce_number=order.get_qr_reference_number(),
            unstructured_message=order.get_unstructured_message(),
            add_cut_info=add_cut_lines,
        )

        return elem

    def __repr__(self):
        return "QR-Invoice"

    def __str__(self):
        return "QR-Invoice"

    def debug(self):
        c = self.canv
        c.setStrokeColor("green")
        c.rect(5 * mm, 5 * mm, 52 * mm, 95 * mm)  # Empfangsschein
        c.rect(67 * mm, 5 * mm, 138 * mm, 95 * mm)  # Zahlteil
        c.rect(67 * mm, 42 * mm, 46 * mm, 46 * mm)  # QR-Code
        c.line(5 * mm, 23 * mm, 57 * mm, 23 * mm)
        c.line(5 * mm, 37 * mm, 57 * mm, 37 * mm)
        c.line(5 * mm, 93 * mm, 57 * mm, 93 * mm)
        c.line(67 * mm, 93 * mm, 118 * mm, 93 * mm)
        c.line(67 * mm, 37 * mm, 118 * mm, 37 * mm)
        c.line(67 * mm, 15 * mm, 205 * mm, 15 * mm)
        c.line(118 * mm, 100 * mm, 118 * mm, 15 * mm)
        c.setStrokeColor("black")

    def get_swiss_qr_payload(self):
        recv = self.payment_receiver
        addr = self.address
        qrpayload = []

        def ln(text=""):
            qrpayload.append(text)

        # QRCH
        # - Header
        # - - QRType
        ln("SPC")
        # - - Version
        ln("0200")
        # - - Coding
        ln("1")

        # - CdtrInf (Empfänger)
        # - - IBAN
        if recv.mode == "QRR":
            ln(recv.qriban.replace(" ", ""))
        else:
            ln(recv.iban.replace(" ", ""))
        # - - Cdtr
        # - - - AdrTp
        ln("K")
        # - - - Name
        ln(recv.invoice_name)
        # - - - StrtNmOrAdrLine1
        ln(recv.invoice_address_1)
        # - - - BldgNbOrAdrLine2
        ln(recv.invoice_address_2)
        # - - - PstCd
        ln()
        # - - - TwnNm
        ln()
        # - - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln(recv.invoice_country)

        # - UltmtCdtr (Entgültiger Zahlungsempfänger)
        # - - AdrTp
        ln()
        # - - Name
        ln()
        # - - StrtNmOrAdrLine1
        ln()
        # - - BldgNbOrAdrLine2
        ln()
        # - - PstCd
        ln()
        # - - TwnNm
        ln()
        # - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln()

        # - CcyAmt
        # - - Amt
        ln(formatprice(self.total))
        # - - Ccy
        ln("CHF")

        # - UltmtDbtr (Entgültiger Zahlungspflichtiger)
        # - - AdrTp
        ln("K")
        # - - Name
        ln(addr["company"] or f"{addr['first_name']} {addr['last_name']}")
        # - - StrtNmOrAdrLine1
        ln(addr["address_1"])
        # - - BldgNbOrAdrLine2
        ln(addr["postcode"] + " " + addr["city"])
        # - - PstCd
        ln()
        # - - TwnNm
        ln()
        # - - Ctry (2-stelliger Landescode gemäss ISO 3166-1)
        ln(addr["country"])

        # - RmtIn
        # - - TP
        ln(recv.mode)
        # - - Ref
        if recv.mode == "QRR":
            ln(self.qr_reference_number.replace(" ", ""))
        else:
            ln()
        # - - AddInf
        # - - - Ustrd
        ln(self.unstructured_message)
        # - - - Trailer
        ln("EPD")
        # - - - StrdBkgInf
        ln(self.billing_information)

        # - AltPmtInf
        # - - AltPmt
        # ln()
        # - - AltPmt
        # ln()

        return "\n".join(qrpayload)

    def draw_qr_invoice(self):
        billing_info = self.billing_information
        ref = self.qr_reference_number
        total = format(self.total, ",.2f").replace(",", " ")
        recv = self.payment_receiver
        addr = self.address

        c = self.canv
        c.saveState()

        # QR-Code

        qrpayload = self.get_swiss_qr_payload()
        qr_code = qr.QrCodeWidget(qrpayload)
        qrbounds = qr_code.getBounds()
        qrwidth = qrbounds[2] - qrbounds[0]
        qrheight = qrbounds[3] - qrbounds[1]
        d = Drawing(
            52.2 * mm,
            52.2 * mm,
            transform=[52.2 * mm / qrwidth, 0, 0, 52.2 * mm / qrheight, 0, 0],
        )  # 46, 46
        d.add(qr_code)
        renderPDF.draw(d, c, 63.9 * mm, 38.9 * mm)  # 67, 42

        # Schweizerkreuz

        c.setFillColor("black")
        c.setStrokeColor("white")
        c.rect(86.5 * mm, 61.5 * mm, 7 * mm, 7 * mm, fill=1, stroke=1)
        c.setFillColor("white")
        c.rect(89.25 * mm, 63 * mm, 1.5 * mm, 4 * mm, fill=1, stroke=0)
        c.rect(88 * mm, 64.25 * mm, 4 * mm, 1.5 * mm, fill=1, stroke=0)

        c.setFillColor("black")
        c.setStrokeColor("black")

        # Begrenzungen Empfangsschein und Zahlteil und Abzutrennen-Hinweis

        if self.add_cut_info:
            c.line(0 * mm, 105 * mm, 210 * mm, 105 * mm)
            c.line(62 * mm, 0 * mm, 62 * mm, 105 * mm)
            c.setFont("Helvetica-Bold", 8)
            c.drawCentredString(
                105 * mm,
                107 * mm,
                pgettext(
                    "QR-Invoice / fixed by SIX group style guide",
                    "Vor der Einzahlung abzutrennen",
                ),
            )

        # Titel

        def titel(t, text, klein=False):
            t.setFont("Helvetica-Bold", 6 if klein else 8)
            t.textLine(text)
            t.moveCursor(0, 2)
            t.setFont("Helvetica", 8 if klein else 10)

        # Empfangsschein Angaben
        t = c.beginText(5 * mm, 90 * mm)
        titel(
            t,
            pgettext(
                "QR-Invoice / fixed by SIX group style guide", "Konto / Zahlbar an"
            ),
            True,
        )
        t.textLine(recv.qriban if recv.mode == "QRR" else recv.iban)
        t.textLine(recv.invoice_name)
        t.textLine(recv.invoice_address_1)
        t.textLine(recv.invoice_address_2)
        t.moveCursor(0, 9)
        if recv.mode == "QRR":
            titel(
                t,
                pgettext("QR-Invoice / fixed by SIX group style guide", "Referenz"),
                True,
            )
            t.textLine(ref)
            t.moveCursor(0, 9)
        titel(
            t,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Zahlbar durch"),
            True,
        )
        t.textLine(addr["company"] or f"{addr['first_name']} {addr['last_name']}")
        t.textLine(addr["address_1"])
        t.textLine(f"{addr['postcode']} {addr['city']}")
        c.drawText(t)

        # Zahlteil Angaben
        t = c.beginText(118 * mm, 97 * mm)
        titel(
            t,
            pgettext(
                "QR-Invoice / fixed by SIX group style guide", "Konto / Zahlbar an"
            ),
        )
        t.textLine(recv.qriban if recv.mode == "QRR" else recv.iban)
        t.textLine(recv.invoice_name)
        t.textLine(recv.invoice_address_1)
        t.textLine(recv.invoice_address_2)
        t.moveCursor(0, 9)
        if recv.mode == "QRR":
            titel(
                t, pgettext("QR-Invoice / fixed by SIX group style guide", "Referenz")
            )
            t.textLine(ref)
            t.moveCursor(0, 9)
        titel(
            t,
            pgettext(
                "QR-Invoice / fixed by SIX group style guide",
                "Zusätzliche Informationen",
            ),
        )
        t.textLine(self.unstructured_message)
        # Note: The billing information was originally printed on the
        #       invoice but is now exclusively included in the QR code.
        #       Reason: It's not human readable and just confusing.
        # t.textLine(billing_info[:len(billing_info)//2)
        # t.textLine(billing_info[len(billing_info)//2:])
        t.moveCursor(0, 9)
        titel(
            t, pgettext("QR-Invoice / fixed by SIX group style guide", "Zahlbar durch")
        )
        t.textLine(addr["company"] or f"{addr['first_name']} {addr['last_name']}")
        t.textLine(addr["address_1"])
        t.textLine(f"{addr['postcode']} {addr['city']}")
        c.drawText(t)

        # Texte
        c.setFont("Helvetica-Bold", 11)
        c.drawString(
            5 * mm,
            97 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Empfangsschein"),
        )
        c.drawString(
            67 * mm,
            97 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Zahlteil"),
        )

        c.setFont("Helvetica-Bold", 6)
        c.drawString(
            5 * mm,
            33 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Währung"),
        )
        c.drawString(
            20 * mm,
            33 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Betrag"),
        )
        c.drawString(
            38 * mm,
            20 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Annahmestelle"),
        )

        c.setFont("Helvetica", 8)
        c.drawString(5 * mm, 30 * mm, "CHF")
        c.drawString(20 * mm, 30 * mm, total)

        c.setFont("Helvetica-Bold", 8)
        c.drawString(
            67 * mm,
            33 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Währung"),
        )
        c.drawString(
            87 * mm,
            33 * mm,
            pgettext("QR-Invoice / fixed by SIX group style guide", "Betrag"),
        )

        c.setFont("Helvetica", 10)
        c.drawString(67 * mm, 29 * mm, "CHF")
        c.drawString(87 * mm, 29 * mm, total)

        # c.setFont("Helvetica-Bold", 7)
        # c.drawString(67*mm, 11*mm, "Name AV1:")
        # c.drawString(67*mm, 8*mm, "Name AV2:")
        #
        # c.setFont("Helvetica", 7)
        # c.drawString(82*mm, 11*mm, "Linie 1")
        # c.drawString(82*mm, 8*mm, "Linie 2")

        # if settings.DEBUG:
        # self.debug()

        c.restoreState()

    def draw(self):
        self.canv.translate(-12 * mm, -12 * mm)
        self.draw_qr_invoice()
